/**
 * \file  bl_main.c
 *
 * \brief Implements main function for StarterWare bootloader
 *
*/

/*
* Copyright (C) 2010-2015 Texas Instruments Incorporated - http://www.ti.com/
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*    Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*    Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*    Neither the name of Texas Instruments Incorporated nor the names of
*    its contributors may be used to endorse or promote products derived
*    from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
*  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
*  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
*  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
*  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
*  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


//#include "uartStdio.h"
#include "bl_copy.h"
#include "bl_platform.h"
#include "bl.h"
#ifdef  XIP_NOR
#include "bl_norexec.h"
#include "soc_AM335x.h"
#endif
#include "hw_types.h"
#include "hw_cm_per.h"
#include "gpio_v2.h"
#include "hw_control_AM335x.h"
#include "hw_cm_wkup.h"

#ifdef ENABLEMMU_CACHE
#include "mmu.h"
#include "cache.h"
#ifdef __TMS470__
#pragma DATA_ALIGN(pageTable, 16384);
static volatile unsigned int pageTable[4*1024];
#endif
#endif


#define APPL_BUILD_VER  "1.0.8"

/*
 *  1.0.8 - Version released with ISDK 1.1.0.5 (Fast Boot)
 *  1.0.7 - Version released with ISDK 1.1.0.3
 *  1.0.6 - Starterware 2.0.1.1 merged
 *  1.0.5 - Version released with ISDK 1.1.0.1
 *  1.0.4 - Version released with ISDK 1.0.0.6
 *  1.0.3 - Version released with ISDK 1.0.0.5
 */

/******************************************************************************
**                    External Variable Declararions 
*******************************************************************************/
extern char *deviceType;
#ifndef FASTBOOT_MODE
extern void bl_UARTPuts(char* buffer);
extern void bl_UARTInit(void);
#else
#define bl_UARTPuts(x) ;
#endif


/******************************************************************************
**                     Local Function Declararion 
*******************************************************************************/

static void (*appEntry)();


/******************************************************************************
**                     Global Variable Definitions
*******************************************************************************/

unsigned int entryPoint;
unsigned int DspEntryPoint;


/******************************************************************************
**                     Global Function Definitions
*******************************************************************************/
#if (defined  XIP_NOR) && (!defined HW_ICE_V2)
extern void* ram_code_load_start;
extern void* ram_code_size;
extern void* ram_code_run_start;

void copy_ram_code()
{
	unsigned int cnt = 0;
	unsigned int len = (unsigned int)&ram_code_size;
	unsigned int * src = (unsigned int *)&ram_code_load_start;
	unsigned int * dst = (unsigned int *)&ram_code_run_start;


	for(cnt = 0 ; cnt < (len/4); cnt++)
	{
		dst[cnt] = src[cnt];
	}
}

#endif
#ifdef ENABLEMMU_CACHE

#define START_ADDR_NOR                     (0x08000000)
#define START_ADDR_DDR                     (0x80000000)
#define START_ADDR_DEV                     (0x44000000)
#define START_ADDR_OCMC                    (0x402f0000)
#define NUM_SECTIONS_DDR                   (512)
#define NUM_SECTIONS_DEV                   (1280)
#define NUM_SECTIONS_OCMC                  (2)
#define NUM_SECTIONS_NOR				   (4)


/*
** Function to setup MMU. This function Maps three regions (1. DDR
** 2. OCMC and 3. Device memory) and enables MMU.
*/
void MMUConfigAndEnable(void)
{
    /*
    ** Define DDR memory region of AM335x. DDR can be configured as Normal
    ** memory with R/W access in user/privileged modes. The cache attributes
    ** specified here are,
    ** Inner - Write through, No Write Allocate
    ** Outer - Write Back, Write Allocate
    */
    REGION regionDdr = {
                        MMU_PGTYPE_SECTION, START_ADDR_DDR, NUM_SECTIONS_DDR,
                        MMU_MEMTYPE_NORMAL_NON_SHAREABLE(MMU_CACHE_WT_NOWA,
                                                         MMU_CACHE_WB_WA),
                        MMU_REGION_NON_SECURE, MMU_AP_PRV_RW_USR_RW,
                        (unsigned int*)pageTable
                       };
    /*
    ** Define OCMC RAM region of AM335x. Same Attributes of DDR region given.
    */
    REGION regionOcmc = {
                         MMU_PGTYPE_SECTION, START_ADDR_OCMC, NUM_SECTIONS_OCMC,
                         MMU_MEMTYPE_NORMAL_NON_SHAREABLE(MMU_CACHE_WT_NOWA,
                                                          MMU_CACHE_WB_WA),
                         MMU_REGION_NON_SECURE, MMU_AP_PRV_RW_USR_RW,
                         (unsigned int*)pageTable
                        };

    /*
    ** Define Device Memory Region. The region between OCMC and DDR is
    ** configured as device memory, with R/W access in user/privileged modes.
    ** Also, the region is marked 'Execute Never'.
    */
    REGION regionDev = {
                        MMU_PGTYPE_SECTION, START_ADDR_DEV, NUM_SECTIONS_DEV,
                        MMU_MEMTYPE_STRONG_ORD_SHAREABLE,
                        MMU_REGION_NON_SECURE,
                        MMU_AP_PRV_RW_USR_RW  | MMU_SECTION_EXEC_NEVER,
                        (unsigned int*)pageTable
                       };
#ifdef FASTBOOT_SPINOR
    /*
    ** Define NOR memory region of AM335x. NOR can be configured as Normal
    ** memory with R/W access in user/privileged modes. The cache attributes
    ** specified here are,
    ** Inner - Write through, No Write Allocate
    ** Outer - Write Back, Write Allocate
    */
    REGION regionNor = {
            MMU_PGTYPE_SECTION, START_ADDR_NOR, NUM_SECTIONS_NOR,
            MMU_MEMTYPE_NORMAL_NON_SHAREABLE(MMU_CACHE_WT_NOWA,
                                             MMU_CACHE_WB_WA),
            MMU_REGION_NON_SECURE, MMU_AP_PRV_RW_USR_RW,
            (unsigned int*)pageTable
                       };
#endif

    /*

    /* Initialize the page table and MMU */
    MMUInit((unsigned int*)pageTable);

    /* Map the defined regions */
    //MMUMemRegionMap(&regionDdr);
    MMUMemRegionMap(&regionOcmc);
    MMUMemRegionMap(&regionDev);
#ifdef XIP_NOR
    MMUMemRegionMap(&regionNor);
#endif

    /* Now Safe to enable MMU */
    MMUEnable((unsigned int*)pageTable);
}


#endif



#ifdef MEASURE_BOOT_TIME
void EnableBootTimeMessurement()
{
        //Powerup GPIO module 
    if( BOOTTIME_GPIO_BASE_ADDR == SOC_GPIO_0_REGS )
    {
        HWREG( SOC_CM_WKUP_REGS + CM_WKUP_GPIO0_CLKCTRL )  |=
                CM_WKUP_GPIO0_CLKCTRL_MODULEMODE_ENABLE;
    }
    if( BOOTTIME_GPIO_BASE_ADDR == SOC_GPIO_1_REGS)
    {
        HWREG( SOC_PRCM_REGS + CM_PER_GPIO1_CLKCTRL )      |=
                CM_PER_GPIO1_CLKCTRL_MODULEMODE_ENABLE;
    }
	if( BOOTTIME_GPIO_BASE_ADDR == SOC_GPIO_2_REGS)
    {
        HWREG( SOC_PRCM_REGS + CM_PER_GPIO2_CLKCTRL )      |=
                CM_PER_GPIO2_CLKCTRL_MODULEMODE_ENABLE;
    }
    if( BOOTTIME_GPIO_BASE_ADDR == SOC_GPIO_3_REGS)
    {
        HWREG( SOC_PRCM_REGS + CM_PER_GPIO3_CLKCTRL )      |=
                CM_PER_GPIO3_CLKCTRL_MODULEMODE_ENABLE;
    }



    GPIOModuleEnable(BOOTTIME_GPIO_BASE_ADDR);
    //Mux the pin
    HWREG( SOC_CONTROL_REGS + CONTROL_CONF_UART_TXD(4)) = (7);
    //Set pin direction
    GPIODirModeSet(BOOTTIME_GPIO_BASE_ADDR,BOOTTIME_GPIO_NUM,GPIO_DIR_OUTPUT);
    //Enable pin
    GPIOPinWrite(BOOTTIME_GPIO_BASE_ADDR,BOOTTIME_GPIO_NUM,GPIO_PIN_LOW);

}
#endif

/*
 * \brief This function initializes the system and copies the image. 
 *
 * \param  none
 *
 * \return none 
*/
int main(void)
{
#ifdef ENABLEMMU_CACHE
    volatile int itr = 0;
    MMUConfigAndEnable();
    CacheEnable(CACHE_ALL);
#endif
#ifdef MEASURE_BOOT_TIME
    EnableBootTimeMessurement();
#endif
    /* Configures PLL and DDR controller*/
    BlPlatformConfig();
    *(unsigned int*)0x44E00508 = 2;			// select 32kHz

    /* UART Initialization */
#ifndef FASTBOOT_MODE
    bl_UARTInit();
#endif
    bl_UARTPuts("\n\r*** StarterWare ");
    //bl_UARTPuts(deviceType);
    bl_UARTPuts(" Boot Loader. Build - ");
    bl_UARTPuts(APPL_BUILD_VER);
    #ifdef HW_ICE_V2
    bl_UARTPuts(" for ICE V2 ");
    #elif defined (HW_ICE)
    bl_UARTPuts(" for ICE V1 ");
    #elif defined (HW_IDK)
    bl_UARTPuts(" for IDK ");
    #endif
    
    #ifdef XIP_NOR
    bl_UARTPuts(" to run from NOR ");
    #endif

#ifdef FASTBOOT_PHY
    mdio_phy_config();
#endif

	//bl_UARTPuts(APPL_BUILD_VER);
#ifdef FASTBOOT_CONFIGNOR
    BlPlatformNORInit();
#endif //FASTBOOT_CONFIGNOR

    /* Copies application from non-volatile flash memory to RAM */

    //bl_UARTPuts("\n\rStart NOR application directly without MMCSDBootCopy!!\n\r");
#ifndef FASTBOOT_SPINOR
    if(ImageCopy() != E_FAIL)
    {
        bl_UARTPuts(" Image Copy Successful, Executing Application..\n\r");

        /* Do any post-copy config before leaving boot loader */
        BlPlatformConfigPostBoot();

        /* Giving control to the application */
        appEntry = (void (*)(void)) entryPoint;
#ifdef MEASURE_BOOT_TIME
        GPIOPinWrite(BOOTTIME_GPIO_BASE_ADDR,BOOTTIME_GPIO_NUM,GPIO_PIN_LOW);
#endif
    #ifdef ENABLEMMU_CACHE
        CacheDataCleanInvalidateAll();
        CacheInstInvalidateAll();

        CacheDisable(CACHE_ALL);
        MMUDisable();
        //for(itr = 0; itr < 1000000 ; itr++);
    #endif
        (*appEntry)( );
    }
#else

#if 0
    while(1)
    {
    	Ethernet_PHY_Register_Dump();
    };
#endif

#ifdef FASTBOOT_XIPNOR
    find_NORImage_from_SPI_exec();
#endif //FASTBOOT_XIPNOR

#endif

#ifdef XIP_NOR
    #if (defined HW_ICE) && (!defined HW_ICE_V2)
    copy_ram_code();
    #endif
    find_NORImage_exec();
#endif
#ifdef MEASURE_BOOT_TIME
    GPIOPinWrite(BOOTTIME_GPIO_BASE_ADDR,BOOTTIME_GPIO_NUM,GPIO_PIN_LOW);
#endif    
    bl_UARTPuts(" Application image not found, Aborting..\n\r");
    BootAbort();

}

void BootAbort(void)
{
#ifdef ENABLEMMU_CACHE
    CacheDataCleanInvalidateAll();
    CacheInstInvalidateAll();

    CacheDisable(CACHE_ALL);
    MMUDisable();
#endif
    while(1);
}

/******************************************************************************
**                              END OF FILE
*******************************************************************************/
